#!/usr/bin/env python3
# Copyright (C) 2019-2020 All rights reserved.
# FILENAME:  load_config.py
# VERSION: 	 0.0.1
# CREATED: 	 2020-11-25 14:35
# AUTHOR: 	 Aekasitt Guruvanich <aekazitt@gmail.com>
# DESCRIPTION:
#
# HISTORY:
#*************************************************************
from typing import Optional, Sequence
from pydantic import BaseModel, validator, StrictBool, StrictInt, StrictStr

class LoadConfig(BaseModel):
  csrf_header_name:Optional[StrictStr] = 'X-CSRF-Token'
  csrf_header_type:Optional[StrictStr] = None
  csrf_in_cookies:Optional[StrictBool] = True
  csrf_methods:Optional[Sequence[StrictStr]] = { 'POST', 'PUT', 'PATCH', 'DELETE' }
  max_age:Optional[StrictInt] = 3600
  secret_key:Optional[StrictStr] = None
  token_locations:Optional[Sequence[StrictStr]] = { 'headers' }
  # In case of using cookies
  cookie_key:Optional[StrictStr] = 'fastapi-csrf-token'
  cookie_path:Optional[StrictStr] = '/'
  cookie_domain:Optional[StrictStr] = None
  cookie_secure:Optional[StrictBool] = False
  cookie_samesite:Optional[StrictStr] = None
  cookie_csrf_protect:Optional[StrictBool] = True
  httponly:Optional[StrictBool] = True

  @validator('csrf_methods', each_item=True)
  def validate_csrf_methods(cls, value):
    if value.upper() not in {'GET', 'HEAD', 'POST', 'PUT', 'DELETE', 'PATCH'}:
      raise ValueError('The "csrf_methods" must be between http request methods')
    return value.upper()

  @validator('cookie_samesite')
  def validate_cookie_samesite(cls, v):
    if value not in ['strict', 'lax', 'none']:
      raise ValueError('The "cookie_samesite" must be between "strict", "lax", "none"')
    return value



#!/usr/bin/env python3
# Copyright (C) 2019-2020 All rights reserved.
# FILENAME:  __init__.py
# VERSION: 	 0.0.1
# CREATED: 	 2020-11-25 14:35
# AUTHOR: 	 Aekasitt Guruvanich <aekazitt@gmail.com>
# DESCRIPTION:
#
# HISTORY:
#*************************************************************
'''
FastAPI extension that provides Csrf Protection Token support
'''

__version__ = '0.0.2'

from .core import CsrfProtect


#!/usr/bin/env python3
# Copyright (C) 2019-2020 All rights reserved.
# FILENAME:  core.py
# VERSION: 	 0.0.1
# CREATED: 	 2020-11-25 14:35
# AUTHOR: 	 Aekasitt Guruvanich <aekazitt@gmail.com>
# DESCRIPTION:
#
# HISTORY:
#*************************************************************
from os import urandom
from hashlib import sha1
from typing import Optional
from fastapi import Request
from fastapi.responses import Response
from itsdangerous import BadData, SignatureExpired, URLSafeTimedSerializer
from fastapi_csrf_protect.csrf_config import CsrfConfig
from fastapi_csrf_protect.exceptions import InvalidHeaderError, MissingTokenError, TokenValidationError

class CsrfProtect(CsrfConfig):
  def __init__(self,req: Request = None, res: Response = None):
    '''
    Retrieve response object if jwt in the cookie
    :param req: all incoming request
    :param res: response from endpoint
    '''
    if res and self.token_in_cookies:
      self._response = res

  def generate_csrf(self, secret_key:Optional[str]=None):
    '''
    Generate a CSRF token. The token is cached for a request, so multiple
    calls to this function will generate the same token.
    '''
    secret_key = secret_key or self._secret_key
    if secret_key is None: raise RuntimeError('A secret key is required to use CSRF.')
    serializer = URLSafeTimedSerializer(secret_key, salt='fastapi-csrf-token')
    token = serializer.dumps(sha1(urandom(64)).hexdigest())
    return token

  def get_csrf_from_headers(self, auth: str) -> str:
    '''
    Get token from the headers
    :param auth: value from HeaderName
    '''
    header_name, header_type = self._csrf_header_name, self._csrf_header_type
    parts = auth.split()
    token = None
    # Make sure the header is in a valid format that we are expecting, ie
    if not header_type:
      # <HeaderName>: <Token>
      if len(parts) != 1:
        raise InvalidHeaderError(f'Bad {header_name} header. Expected value "<Token>"')
      token = parts[0]
    else:
      # <HeaderName>: <HeaderType> <Token>
      if not re.match(r"{}\s".format(header_type),auth) or len(parts) != 2:
        raise InvalidHeaderError(f'Bad {header_name} header. Expected value "{header_type} <Token>"')
      token = parts[1]
    return token # TODO Validate

  def set_csrf_cookie(self, response: Optional[Response]= None) -> None:
    if response and not isinstance(response,Response):
      raise TypeError('The response must be an object response FastAPI')
    response = response or self._response
    response.set_cookie(
      self._cookie_key,
      self.generate_csrf(self._secret_key),
      max_age=self._max_age,
      path=self._cookie_path,
      domain=self._cookie_domain,
      secure=self._cookie_secure,
      httponly=self._httponly,
      samesite=self._cookie_samesite
    )

  def unset_csrf_cookie(self, response: Optional[Response]= None) -> None:
    '''
    Remove Csrf Protection token from the response cookies
    :param response: The FastAPI response object to delete the access cookies in.
    '''
    if response and not isinstance(response,Response):
      raise TypeError('The response must be an object response FastAPI')
    response = response or self._response
    response.delete_cookie(
      self._cookie_key,
      path=self._cookie_path,
      domain=self._cookie_domain
    )

  def validate_csrf(self, data, secret_key=None, time_limit=None):
    '''
    Check if the given data is a valid CSRF token. This compares the given
    signed token to the one stored in the session.
    :param data: The signed CSRF token to be checked.
    :param secret_key: Used to securely sign the token.
        Default is set in CsrfConfig
    :param time_limit: Number of seconds that the token is valid.
        Default is set in CsrfConfig
    :raises TokenValidationError: Contains the reason that validation failed.
    '''
    secret_key = secret_key or self._secret_key
    time_limit = time_limit or self._max_age
    if not data:
      raise TokenValidationError('The CSRF token is missing.')
    serializer = URLSafeTimedSerializer(secret_key, salt='fastapi-csrf-token')
    try:
      token = serializer.loads(data, max_age=time_limit)
    except SignatureExpired:
      raise TokenValidationError('The CSRF token has expired.')
    except BadData:
      raise TokenValidationError('The CSRF token is invalid.')

  def validate_csrf_in_cookies(self, request: Request, secret_key:Optional[str]=None, field_name:Optional[str]=None):
    secret_key = secret_key or self._secret_key
    if secret_key is None: raise RuntimeError('A secret key is required to use CSRF.')
    field_name = field_name or self._cookie_key
    cookie = request.cookies.get(field_name)
    if cookie is None: raise MissingTokenError(f'Missing Cookie {field_name}')
    self.validate_csrf(cookie)



#!/usr/bin/env python3
# Copyright (C) 2019-2020 All rights reserved.
# FILENAME:  exceptions.py
# VERSION: 	 0.0.1
# CREATED: 	 2020-11-25 14:35
# AUTHOR: 	 Aekasitt Guruvanich <aekazitt@gmail.com>
# DESCRIPTION:
#
# HISTORY:
#*************************************************************
class CsrfProtectError(Exception):
  def __init__(self, status_code, message):
    self.status_code = status_code
    self.message = message

class InvalidHeaderError(CsrfProtectError):
  def __init__(self, message):
    super().__init__(422, message)

class MissingTokenError(CsrfProtectError):
  def __init__(self, message):
    super().__init__(400, message)

class TokenValidationError(CsrfProtectError):
  def __init__(self, message):
    super().__init__(401, message)



#!/usr/bin/env python3
# Copyright (C) 2019-2020 All rights reserved.
# FILENAME:  setup.py
# VERSION: 	 0.0.1
# CREATED: 	 2020-11-25 14:35
# AUTHOR: 	 Aekasitt Guruvanich <aekazitt@gmail.com>
# DESCRIPTION:
#
# HISTORY:
#*************************************************************
from setuptools import setup

# Load Long Description from README
long_description = ''
with open('README.md', 'r') as readme:
  long_description = readme.read()

setup(
  name='fastapi-csrf-protect',
  version='0.0.2',
  description='Simple integration of Cross-Site Request Forgery (XSRF) Protection by using either Cookies or Context combined with Headers',
  py_modules=['fastapi_csrf_protect'],
  package_dir={'': 'fastapi_csrf_protect'},
  classifiers=[
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8'
  ],
  long_description=long_description,
  long_description_content_type='text/markdown',
  install_requires=[
    'fastapi ~= 0.61.2',
    'itsdangerous ~= 1.1.0',
    'pydantic ~= 1.7.2'
  ],
  extras_require={
    'dev': [
      'pytest >= 3.7'
    ]
  }
)


#!/usr/bin/env python3
# Copyright (C) 2019-2020 All rights reserved.
# FILENAME:  fastapi_csrf_config.py
# VERSION: 	 0.0.1
# CREATED: 	 2020-11-25 14:35
# AUTHOR: 	 Aekasitt Guruvanich <aekazitt@gmail.com>
# DESCRIPTION:
#
# HISTORY:
#*************************************************************
from typing import Callable, List, Sequence
from pydantic import ValidationError
from fastapi_csrf_protect.load_config import LoadConfig

class CsrfConfig(object):
  _csrf_header_name:str = 'X-CSRF-Token'
  _csrf_header_type:str = None
  _csrf_in_cookies:bool = True
  _csrf_methods:Sequence[str] = { 'POST', 'PUT', 'PATCH', 'DELETE' }
  _max_age:int = 3600
  _secret_key:str = None
  _token_locations:Sequence[str] = { 'headers' }
  # In case of using cookies
  _cookie_key:str = 'fastapi-csrf-token'
  _cookie_path:str = '/'
  _cookie_domain:str = None
  _cookie_secure:bool = False
  _cookie_samesite:bool = None
  _cookie_csrf_protect:bool = True
  _httponly:bool = True

  @property
  def token_in_headers(self) -> bool:
    return 'headers' in self._token_locations

  @property
  def token_in_cookies(self) -> bool:
    return 'cookies' in self._token_locations

  @classmethod
  def load_config(cls, settings: Callable[..., List[tuple]]) -> 'CsrfConfig':
    try:
      config = LoadConfig(**{key.lower():value for key,value in settings()})
      cls._csrf_header_name = config.csrf_header_name
      cls._csrf_header_type = config.csrf_header_type
      cls._csrf_in_cookies = config.csrf_in_cookies
      cls._csrf_methods = config.csrf_methods
      cls._max_age = config.max_age
      cls._secret_key = config.secret_key
      cls._token_locations = config.token_locations
      cls._cookie_key = config.cookie_key
      cls._cookie_path = config.cookie_path
      cls._cookie_domain = config.cookie_domain
      cls._cookie_secure = config.cookie_secure
      cls._cookie_samesite = config.cookie_samesite
      cls._cookie_csrf_protect = config.cookie_csrf_protect
      cls._httponly = config.httponly
    except ValidationError: raise
    except Exception:
      raise TypeError('CsrfConfig must be pydantic "BaseSettings" or list of tuple')